from django.contrib import messages
from django.core.exceptions import ValidationError
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.generic import DetailView, ListView

from servala.core.models import (
    ControlPlaneCRD,
    Service,
    ServiceInstance,
    ServiceOffering,
)
from servala.frontend.forms.service import (
    ControlPlaneSelectForm,
    ServiceFilterForm,
    ServiceInstanceFilterForm,
)
from servala.frontend.views.mixins import HtmxViewMixin, OrganizationViewMixin


class ServiceListView(OrganizationViewMixin, ListView):
    """View to display all available services for an organization."""

    template_name = "frontend/organizations/services.html"
    context_object_name = "services"
    model = Service
    permission_type = "view"

    def get_queryset(self):
        """Return all services."""
        services = Service.objects.all().select_related("category")
        if self.filter_form.is_valid():
            services = self.filter_form.filter_queryset(services)
        return services

    @cached_property
    def filter_form(self):
        return ServiceFilterForm(data=self.request.GET or None)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["filter_form"] = self.filter_form
        return context


class ServiceDetailView(OrganizationViewMixin, DetailView):
    template_name = "frontend/organizations/service_detail.html"
    context_object_name = "service"
    model = Service
    permission_type = "view"

    def get_queryset(self):
        return Service.objects.select_related("category").prefetch_related(
            "offerings",
            "offerings__provider",
        )


class ServiceOfferingDetailView(OrganizationViewMixin, HtmxViewMixin, DetailView):
    template_name = "frontend/organizations/service_offering_detail.html"
    context_object_name = "offering"
    model = ServiceOffering
    permission_type = "view"
    fragments = ("service-form",)

    def has_permission(self):
        return self.has_organization_permission()

    def get_queryset(self):
        return ServiceOffering.objects.all().select_related(
            "service", "service__category", "provider"
        )

    @cached_property
    def planes(self):
        return self.object.control_planes.all()

    @cached_property
    def select_form(self):
        data = None
        if "control_plane" in self.request.GET:
            data = self.request.GET
        elif self.request.method == "POST" and self.context_object:
            data = {"control_plane": self.context_object.control_plane_id}
        return ControlPlaneSelectForm(data=data, planes=self.planes)

    @cached_property
    def selected_plane(self):
        if self.select_form.data and self.select_form.is_valid():
            return self.select_form.cleaned_data["control_plane"]
        field = self.select_form.fields["control_plane"]
        return field.initial or field.queryset.first()

    @cached_property
    def context_object(self):
        if self.request.method == "POST":
            return ControlPlaneCRD.objects.filter(
                pk=self.request.POST.get("context"),
                # Make sure we don’t use a malicious ID
                control_plane__in=self.planes,
            ).first()
        return ControlPlaneCRD.objects.filter(
            control_plane=self.selected_plane, service_offering=self.object
        ).first()

    def get_instance_form(self):
        return self.context_object.model_form_class(
            data=self.request.POST if self.request.method == "POST" else None,
            initial={
                "organization": self.request.organization,
                "context": self.context_object,
            },
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["select_form"] = self.select_form
        context["has_control_planes"] = self.planes.exists()
        context["selected_plane"] = self.selected_plane
        context["service_form"] = self.get_instance_form()
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)

        if not self.context_object:
            context["form_error"] = True
            return self.render_to_response(context)

        form = self.get_instance_form()
        if form.is_valid():
            try:
                service_instance = ServiceInstance.create_instance(
                    organization=self.organization,
                    name=form.cleaned_data["name"],
                    context=self.context_object,
                    created_by=request.user,
                    spec_data=form.get_nested_data().get("spec"),
                )
                return redirect(service_instance.urls.base)
            except ValidationError as e:
                messages.error(self.request, e.message or str(e))
            except Exception as e:
                messages.error(self.request, str(e))

        # If the form is not valid or if the service creation failed, we render it again
        context["service_form"] = form
        return self.render_to_response(context)


class ServiceInstanceDetailView(OrganizationViewMixin, DetailView):
    """View to display details of a specific service instance."""

    template_name = "frontend/organizations/service_instance_detail.html"
    context_object_name = "instance"
    model = ServiceInstance
    permission_type = "view"
    slug_field = "name"

    def get_queryset(self):
        """Return service instance for the current organization."""
        return ServiceInstance.objects.filter(
            organization=self.request.organization
        ).select_related(
            "context__service_offering__provider",
            "context__control_plane",
            "context__service_definition__service",
        )


class ServiceInstanceListView(OrganizationViewMixin, ListView):
    template_name = "frontend/organizations/service_instances.html"
    context_object_name = "instances"
    model = ServiceInstance
    permission_type = "view"

    @cached_property
    def filter_form(self):
        return ServiceInstanceFilterForm(data=self.request.GET or None)

    def get_queryset(self):
        """Return all service instances for the current organization with filtering."""
        queryset = ServiceInstance.objects.filter(
            organization=self.request.organization
        )
        if self.filter_form.is_valid():
            queryset = self.filter_form.filter_queryset(queryset)
        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["organization"] = self.request.organization
        context["filter_form"] = self.filter_form
        return context
